/*
 * $Id: hdn_crypto.c,v 1.6 2003/02/04 06:54:13 xvr Exp $
 * Created: 09/03/2002
 *
 * xvr (c) 2002
 * xvr@crazyboy.com
 */

#include "hdn_crypto.h"

static unsigned char iv[] = {'x', 'v', 'r', 'j', 'Z', ':', 'y', 'x'};

unsigned char *hdn_hash (char *in)
{
    unsigned char *digest;
    EVP_MD_CTX ctx;

    if (!in) 
	return NULL;

    if (!(digest = malloc (EVP_MAX_MD_SIZE)))
	return NULL;

    EVP_DigestInit (&ctx, EVP_sha1());
    EVP_DigestUpdate (&ctx, in, strlen (in));
    EVP_DigestFinal (&ctx, digest, NULL);

    return digest;
}

void hdn_encrypt (struct hdn_data **inout, unsigned char *key)
{
    unsigned char *out = 0;
    int  out_sz, sz, total_sz;
    struct hdn_data *in = (*inout);
    EVP_CIPHER_CTX ctx;

    /*
     * init context, bf in cbc mode, default impl
     */
    EVP_EncryptInit (&ctx, EVP_bf_cbc(), key, iv);
    out = malloc (sizeof (in->sz) + in->sz + EVP_CIPHER_CTX_block_size(&ctx));

    if (!out)
    {
	fprintf (stderr, "error allocating memory for encryption\n");
	exit (1);
    }

    /*
     * save the size, and make sure that it's a multiple of the cipher
     * block size
     */
    sz = in->sz;
    in->sz += EVP_CIPHER_CTX_block_size(&ctx) - ((in->sz + sizeof (in->sz)) % EVP_CIPHER_CTX_block_size(&ctx));
    
    /*
     * whiten it
     */
    hdn_xor (&in->sz, hdn_hash (key), sizeof (in->sz));

    /*
     * encrypt everything
     */
    EVP_EncryptUpdate (&ctx, out, &out_sz, (char *)in, sizeof(in->sz) + sz);
    total_sz = out_sz;

    EVP_EncryptFinal (&ctx, out + total_sz, &out_sz);
    total_sz += out_sz;

    /*
     * store this gunk
     */
    (*inout) = realloc (*inout, sizeof (struct hdn_data) + total_sz);

    if (!(*inout))
    {
	fprintf (stderr, "error allocating memory for encryption\n");
	exit (1);
    }

    (*inout)->sz = total_sz;
    memcpy ((*inout)->content, out, total_sz);
    
    /*
     * cleanup
     */
    EVP_CIPHER_CTX_cleanup (&ctx);
    if (out) free (out);
}

void hdn_decrypt (struct hdn_data **inout, unsigned char *key)
{
    unsigned char *out = 0;
    int  out_sz;
    struct hdn_data *in = (*inout);
    EVP_CIPHER_CTX ctx;

    /*
     * init
     */
    EVP_DecryptInit (&ctx, EVP_bf_cbc(), key, iv);
    out = malloc (in->sz + EVP_CIPHER_CTX_block_size(&ctx));

    if (!out)
    {
	fprintf (stderr, "error allocating memory for decryption\n");
	exit (1);
    }

    /*
     * decrypt
     */
    EVP_DecryptUpdate (&ctx, out, &out_sz, in->content, in->sz);
    EVP_DecryptFinal (&ctx, out + out_sz, &out_sz);

    /*
     * store only the right length worth of decryption
     */    
    in = (struct hdn_data *)out;
    hdn_xor (&in->sz, hdn_hash (key), sizeof (in->sz));
    (*inout) = realloc (*inout, in->sz + sizeof (struct hdn_data));

    if (!(*inout))
    {
	fprintf (stderr, "error allocating memory for decryption\n");
	exit (1);
    }

    memcpy (*inout, out, in->sz + sizeof (struct hdn_data));

    /*
     * cleanup
     */
    EVP_CIPHER_CTX_cleanup (&ctx);
    if (out) free (out);    
}

