/*
 * $Id: hdn_decode.c,v 1.11 2003/02/04 06:54:13 xvr Exp $
 * Created: 08/21/2002
 *
 * xvr (c) 2002-2003
 * xvr@crazyboy.com
 */

#include "hdn_decode.h"

static void hdn_usage (void)
{
    fprintf (stderr,
	     "Usage:\n"
	     "\thydan-decode filename\n");
    exit (1);
}

/*
 * decode the message
 */
static 
void hdn_decode_message (struct hdn_host_section *host_sections,
			 struct hdn_data *mesg_data,
			 unsigned char *key)
{
    off_t mesg_curr_pos = 0;
    char  mesg_curr_bit = 0;
    char is_sz_extracted = 0;
    char *addr = NULL;
    struct instr insn;

    /*
     * keep on going while we have a section and data to read.
     */
    while ( (addr = hdn_get_next_address (host_sections)) && 
	    (mesg_curr_pos < mesg_data->sz))
    {
	bzero (&insn, sizeof insn);
	disassemble_address (addr, &insn);

	/*
	 * if we can decode this instruction, do it.
	 */
	if (hdn_is_valid_insn (&insn, addr))
	{
	    int i, bitval, bits = 0;

	    //get the bit value encoded, and the number of bits
	    bitval = hdn_insn_val (&insn, addr, &bits);

#if 0
	    if (bits)
		fprintf (stderr, "%02x%02x: %s %s, %s\t%d\t%d\n",
			 (unsigned char)*(addr),
			 (unsigned char)*(addr + 1),
			 insn.mnemonic, insn.dest, insn.src, 
			 bits, mesg_curr_pos);
#endif

	    //set each bit appropriately at destination
	    for (i = 0; (i < bits) && (mesg_curr_pos < mesg_data->sz); i++)
	    {
		//if ith bit out of bits is set
		if ((bitval << (8 - bits + i)) & 128)
		    *(mesg_data->content + mesg_curr_pos) |= 
			128 >> mesg_curr_bit;

		mesg_curr_bit++;
		mesg_curr_bit %= 8;
		if (!mesg_curr_bit) mesg_curr_pos++;
	    }

	    /* 
	     * if size of message has been extracted, decrypt it, and
	     * use it to bound further extraction
	     */
	    if (!is_sz_extracted && 
		mesg_curr_pos > sizeof (mesg_data->sz) + 8)
	    {
		struct hdn_data *tmp;

		tmp = malloc (8 + sizeof (tmp->sz) + sizeof (struct hdn_data));
		tmp->sz = 8 + sizeof (tmp->sz);
		memcpy (&tmp->content, mesg_data->content, 8 + sizeof (tmp->sz));

		hdn_decrypt (&tmp, key);
		is_sz_extracted = 1;
		mesg_data->sz = tmp->sz + sizeof (tmp->sz);
	    }
	}
    }
}

int hdn_decode_main (int argc, char **argv)
{
    unsigned char   *password;
    struct hdn_data *host_data = 0, *mesg_data = 0;
    struct hdn_host_section *code_sections, *curr_section;

    /*
     * get application data
     */
    if      (argc == 1) host_data = hdn_fdread (STDIN_FILENO);
    else if (argc == 2) host_data = hdn_fileread (argv[1]);
    else                hdn_usage ();

    if (!host_data)
	goto out;

    disassemble_init (0, INTEL_SYNTAX);

    /*
     * get the code segments
     */
    if (!(code_sections = hdn_get_code_locs (host_data->content)))
    {
	fprintf (stderr, "Error extracting code segments from host file\n");
	goto out;
    }	

    /*
     * get all possible data, host_data->sz is the upper bound of
     * available data in the file
     */
    mesg_data = malloc (sizeof (struct hdn_data) + host_data->sz);

    if (!mesg_data)
    {
	perror ("malloc");
	goto out;
    }

    mesg_data->sz = host_data->sz;
    bzero (mesg_data->content, mesg_data->sz);

    /*
     * get password and seed the random number generator
     */
    password = getpass ("Password: ");
    hdn_srandom (password);
    hdn_decode_message (code_sections, mesg_data, password);

    /*
     * decrypt it
     */
    hdn_decrypt (&mesg_data, password);

    bzero (password, _PASSWORD_LEN);

    /*
     * out you go
     */
    hdn_fdwrite (STDOUT_FILENO, mesg_data);

  out:
    if (host_data) free (host_data);
    if (mesg_data) free (mesg_data);
    host_data = NULL;
    mesg_data = NULL;
    while (code_sections) 
    {
	curr_section = code_sections->next;
	free (code_sections);
	code_sections = curr_section;
    }
    disassemble_cleanup ();
    return 0;
}

