/*
 * $Id: hdn_insns.c,v 1.12 2003/02/01 18:03:12 xvr Exp $
 * Created: 08/23/2002
 *
 * xvr (c) 2002
 * xvr@crazyboy.com
 */

#include "hdn_insns.h"

/*
 * the following macros allow to easily test the values in the mod/rm
 * byte in the instruction, which is structured as follows:
 *   00 000 000
 * mod |reg| rm
 */
#define MOD_VAL(modrmb) ( (modrmb & 192) >> 6)
#define REG_VAL(modrmb) ( (modrmb &  56) >> 3)
#define RM_VAL(modrmb)  ( (modrmb &   7)     )

/*
 * defines the relationship between the *_op fields in the
 * instruction structure and the value itself
 */
#define NU 0x00   /* Field Not Used                     */
#define EQ 0x01   /* Field should be equal to the value */
#define GT 0x02   /* Greater than the value             */
#define GE 0x03   /* Greater than or equal to the value */
#define LT 0x04   /* Less than the value                */
#define LE 0x05   /* Less than or equal to the value    */
#define RR 0x06   /* Special case: REG_VAL == RM_VAL    */
#define NEG 0x07  /* Special case: always match, return the negative */

/*
 * description of an instruction that can be substituted
 */
struct sinstr 
{    
    char opcd;         /* its opcode */
    
    char modval_op;
    char modval;       /* mod value */

    char regval_op;  
    char regval;       /* the reg value in the mod/rm byte */
    
    char rmval_op;
    char rmval;        /* the rm value in the mod/rm byte */
    
    char immval_op;
    long long immval; /* the immediate value */

    char *desc;       /* instruction description */
    unsigned int : 0;  /* pad to the next word boundary */
};

#define SINSTR_END { -1, -1, -1, -1, -1, -1, -1, -1, -1, NULL}

struct sinstr toac8_table[] =
{
    { 0x84, NU, -1, RR, -1, RR, -1, NU, -1, "test r8 , r8" },
    { 0x08, NU, -1, RR, -1, RR, -1, NU, -1, "or   r8 , r8" },
    { 0x0A, NU, -1, RR, -1, RR, -1, NU, -1, "or*  r8 , r8" },
    { 0x20, NU, -1, RR, -1, RR, -1, NU, -1, "and  r8 , r8" },
    { 0x22, NU, -1, RR, -1, RR, -1, NU, -1, "and* r8 , r8" },
    SINSTR_END, /* end of struct */
};

struct sinstr toac32_table[] =
{
    { 0x85, NU, -1, RR, -1, RR, -1, NU, -1, "test r32, r32" },
    { 0x09, NU, -1, RR, -1, RR, -1, NU, -1, "or   r32, r32" },
    { 0x0B, NU, -1, RR, -1, RR, -1, NU, -1, "or*  r32, r32" },
    { 0x21, NU, -1, RR, -1, RR, -1, NU, -1, "and  r32, r32" },
    { 0x23, NU, -1, RR, -1, RR, -1, NU, -1, "and* r32, r32" },
    SINSTR_END, /* end of struct */
};

struct sinstr cmp8_table[] =
{
    { 0x38, NU, -1, RR, -1, RR, -1, NU, -1, "cmp  r8 , r8" },
    { 0x3A, NU, -1, RR, -1, RR, -1, NU, -1, "cmp* r8 , r8" },
    SINSTR_END, /* end of struct */
};

struct sinstr cmp32_table[] =
{
    { 0x39, NU, -1, RR, -1, RR, -1, NU, -1, "cmp  r32, r32" },
    { 0x3B, NU, -1, RR, -1, RR, -1, NU, -1, "cmp* r32, r32" },
    SINSTR_END, /* end of struct */
};

struct sinstr toasxc8_table[] =
{
    { 0xA8, NU, -1, NU, -1, NU, -1, EQ, -1, "test al , -1" },
    { 0x0C, NU, -1, NU, -1, NU, -1, EQ,  0, "or   al ,  0" },
    { 0x24, NU, -1, NU, -1, NU, -1, EQ, -1, "and  al , -1" },
    { 0x04, NU, -1, NU, -1, NU, -1, EQ,  0, "add  al ,  0" },
    { 0x2C, NU, -1, NU, -1, NU, -1, EQ,  0, "sub  al ,  0" },
    { 0x34, NU, -1, NU, -1, NU, -1, EQ,  0, "xor  al ,  0" },
    { 0x3C, NU, -1, NU, -1, NU, -1, EQ,  0, "cmp  al ,  0" },
    SINSTR_END, /* end of struct */
};

struct sinstr toasxc32_table[] =
{
    { 0xA9, NU, -1, NU, -1, NU, -1, EQ, -1, "test eax, -1" },
    { 0x0D, NU, -1, NU, -1, NU, -1, EQ,  0, "or   eax,  0" },
    { 0x25, NU, -1, NU, -1, NU, -1, EQ, -1, "and  eax, -1" },
    { 0x05, NU, -1, NU, -1, NU, -1, EQ,  0, "add  eax,  0" },
    { 0x2D, NU, -1, NU, -1, NU, -1, EQ,  0, "sub  eax,  0" },
    { 0x35, NU, -1, NU, -1, NU, -1, EQ,  0, "xor  eax,  0" },
    { 0x3D, NU, -1, NU, -1, NU, -1, EQ,  0, "cmp  eax,  0" },
    SINSTR_END, /* end of struct */
};

struct sinstr addsub8_table[] =
{
    { 0x04, NU, -1, NU, -1, NU, -1, NEG, -1, "add  al, imm8" },
    { 0x2C, NU, -1, NU, -1, NU, -1, NEG, -1, "sub  al, imm8" },
    SINSTR_END, /* end of struct */
};

struct sinstr addsub8_table2[] = 
{ 
    { 0x80, NU, -1, EQ,  0, NU, -1, NEG, -1, "add r/m8, imm8" },
    { 0x80, NU, -1, EQ,  5, NU, -1, NEG, -1, "sub r/m8, imm8" },
    SINSTR_END, /* end of struct */
};

struct sinstr addsub32_table[] =
{
    { 0x05, NU, -1, NU, -1, NU, -1, NEG, -1, "add eax, imm32" },
    { 0x2D, NU, -1, NU, -1, NU, -1, NEG, -1, "sub eax, imm32" },
    SINSTR_END, /* end of struct */
};

struct sinstr addsub32_table2[] = 
{ 
    { 0x81, NU, -1, EQ,  0, NU, -1, NEG, -1, "add r/m32, imm32" },
    { 0x81, NU, -1, EQ,  5, NU, -1, NEG, -1, "sub r/m32, imm32" },
    SINSTR_END, /* end of struct */
};

struct sinstr addsub32_table3[] = 
{ 
    { 0x83, NU, -1, EQ,  0, NU, -1, NEG, -1, "add r/m32, imm8" },
    { 0x83, NU, -1, EQ,  5, NU, -1, NEG, -1, "sub r/m32, imm8" },
    SINSTR_END, /* end of struct */
};

/* only works on registers, since result is stored in destination  */
struct sinstr xor32_table1[] = 
{
    { 0x31, EQ, 3, RR, -1, RR, -1, NU, -1, "xor r/m32, r32"   },
    { 0x33, EQ, 3, RR, -1, RR, -1, NU, -1, "xor r32  , r/m32" },
    { 0x29, EQ, 3, RR, -1, RR, -1, NU, -1, "sub r/m32, r32"   },
    { 0x2B, EQ, 3, RR, -1, RR, -1, NU, -1, "sub r32  , r/m32" },    
    SINSTR_END, /* end of struct */
};

struct sinstr_desc 
{
    char *desc;
    struct sinstr *table;
};

struct sinstr_desc sinstr_table[] = 
{
    {"toac8"     , toac8_table },
    {"toac32"    , toac32_table },
    {"cmp8"      , cmp8_table },
    {"cmp32"     , cmp32_table },
    {"toasxc8"   , toasxc8_table},
    {"toasxc32"  , toasxc32_table},
    {"addsub8"   , addsub8_table},
    {"addsub8-2" , addsub8_table2},
    {"addsub32-1", addsub32_table },
    {"addsub32-2", addsub32_table2 },
    {"addsub32-3", addsub32_table3 },
    {"xor32_table1", xor32_table1 },
    {NULL, NULL},
};


/*
 * compare orig and new using the supplied operation
 */
static int hdn_test_val (char operation, long long orig, long long new)
{
    switch (operation)
    {
	case EQ: return (orig == new);
	case GT: return (orig >  new);
	case GE: return (orig >= new);
	case LT: return (orig <  new);
	case LE: return (orig <= new);

	case NEG:
	case RR: //handled elsewhere
	case NU: //skip
	    return 1; 
    }

    return 0;
}

/*
 * returns the size of the source
 */
static int hdn_src_sz (struct instr *insn)
{
    switch (insn->srcType & OP_SIZE_MASK)
    {
	case OP_BYTE:  return 1;
	case OP_HWORD: return 2;
	case OP_WORD:  return 4;
	case OP_DWORD: return 8;
    }

    return 0;
}

/*
 * returns the immediate value
 */
static long long hdn_get_imm_val (struct instr *insn, unsigned char *host)
{
    int sz;
    long long imm = 0;

    sz = hdn_src_sz (insn);

    memcpy (&imm, host + insn->size - sz, sz);
    return imm;
}

/*
 * return only the first sz bytes of n
 */
static long long hdn_truncate_number (struct instr *insn, long long n)
{
    int temp = n;
    int sz = hdn_src_sz (insn);

    n = 0;
    memcpy (&n, &temp, sz);

    return n;
}

/*
 * set the immediate value
 */
static void hdn_set_imm_val (struct instr *insn, unsigned char *host, 
			     long long imm)
{
    int sz = hdn_src_sz (insn);

    memcpy (host + insn->size - sz, &imm, sz);
}

/*
 * does the instruction encoded by opcd,mdrm match what's in instr?
 */
static int hdn_test_insn (char opcd, char mdrm, struct sinstr *instr, 
			  struct instr *insn,
			  unsigned char *host)
{
    /* test the opcode */
    if (!hdn_test_val (EQ, instr->opcd, opcd)) return 0;
	    
    /* lookout for special RR case */
    if (instr->regval_op == RR || instr->rmval_op == RR)
	if (!hdn_test_val (EQ, REG_VAL(mdrm), RM_VAL(mdrm)))
	    return 0;
	
    /* test the reg val */
    if (!hdn_test_val (instr->regval_op,
		       instr->regval, 
		       REG_VAL(mdrm))) return 0;

    /* test the rm val */
    if (!hdn_test_val (instr->rmval_op,
		       instr->rmval,
		       RM_VAL(mdrm))) return 0;

    /* immediate value */
    if (!hdn_test_val (instr->immval_op,
		       hdn_truncate_number (insn, instr->immval), 
		       hdn_get_imm_val (insn, host))) return 0;

    return 1;
}

/*
 * returns the number of instructions we have in the given set.
 */
static int hdn_num_insns (struct sinstr *table)
{
    int i = 0;
    struct sinstr end = SINSTR_END;

    while (memcmp (&table[i], &end, sizeof (struct sinstr)))
	i++;

    return i;
}

/*
 * same as below, except returns the index of the set of instructions
 * where the instruction was found.
 */
static int _hdn_is_valid_insn (struct instr *insn, 
			       unsigned char *host, int val, int *insn_set)
{
    int set, ins, num_insns, tmp_bits, max_bits = 0;
    unsigned char opcd = host[0];
    unsigned char mdrm = host[1];

    for (set = 0; sinstr_table[set].table; set++)
    {
	num_insns = hdn_num_insns (sinstr_table[set].table);

	/*
	 * If val is specified, see if we can express it using this
	 * set of instructions.  else, don't waste time here.
	 */
	if (val >= 0 && val > num_insns) continue;

	for (ins = 0; ins < num_insns; ins++)
	{
	    //have we found the instruction?
	    if (!hdn_test_insn (opcd, mdrm, &sinstr_table[set].table[ins], insn,host))
		continue;

	    /*
	     * have to use our version of log, since log () rounds too
	     * much, and log1p doesn't work consistently accross
	     * platforms.
	     */
	    tmp_bits = hdn_log2 (num_insns);

#ifdef VARBITS	    
	    /*
	     * if we have more than 2 more instructions than the
	     * number of bits that we can fully express, use the last
	     * instruction as an 'invalid' marker, and encode one more
	     * bit than we normally could.  Assuming random
	     * distribution in 0s and 1s, this means that at the worst
	     * case, half of the instructions will not get used --
	     * which is equivalent to using the lesser bit encoding
	     * scheme.  But in the better cases, we can gain from
	     * this, despite the waste.
	     */
	    if ((num_insns - pow (2, tmp_bits)) > 2)
	    {
		//this the invalid marker, don't use to encode
		if (ins == (num_insns - 1))
		    continue;

		tmp_bits++;
	    }
#endif
	    if (tmp_bits > max_bits)
	    {
		if (insn_set) *insn_set = set;
		max_bits = tmp_bits;
	    }
	}
    }

    return max_bits;
}

/*
 * returns the max number of bits that can be encoded, 0 if invalid
 * instruction.
 */
int hdn_is_valid_insn (struct instr *insn, unsigned char *host)
{
    return _hdn_is_valid_insn (insn, host, -1, NULL);
}

/*
 * returns the description of the encoded instruction
 */
char *hdn_insn_desc (struct instr *insn, unsigned char *host, char **insn_desc)
{
    char *ret = NULL;
    int insn_set, bits;

    if (insn_desc)
	*insn_desc = NULL;

    if (!(bits = _hdn_is_valid_insn (insn, host, -1, &insn_set))) 
	goto out;

    ret = sinstr_table[insn_set].desc;
    if (insn_desc)
	*insn_desc = sinstr_table[insn_set].table[
	    hdn_insn_val (insn, host, NULL)].desc;

  out:
    return ret;
}

int hdn_is_neg_insn (struct instr *insn, unsigned char *host)
{
    int bits, i;
    int insn_set, num_insns;
    unsigned char opcd = host[0];
    unsigned char mdrm = host[1];

    if (!(bits = _hdn_is_valid_insn (insn, host, -1, &insn_set)))
	return 0;

    num_insns = hdn_num_insns (sinstr_table[insn_set].table);
    
    for (i = 0; i < num_insns; i++)
    {
	/*
	 * look for the instruction
	 */
	if (hdn_test_insn (opcd, mdrm, &sinstr_table[insn_set].table[i],
			   insn, host))
	{
	    if (sinstr_table[insn_set].table[i].immval_op == NEG &&
		strtol (insn->src, NULL, 0) < 0)
		return 1;
	}
    }

    return 0;
}

/*
 * return the value of the encoded instruction.
 */
int hdn_insn_val (struct instr *insn, unsigned char *host, int *numbits)
{
    int ret = 0;
    int insn_set, bits, num_insns;
    unsigned char opcd = host[0];
    unsigned char mdrm = host[1];

    /* get number of bits that can be encoded */
    if (!(bits = _hdn_is_valid_insn (insn, host, -1, &insn_set))) goto out;
    num_insns = hdn_num_insns (sinstr_table[insn_set].table);
    
    for (ret = 0; ret < num_insns; ret++)
    {
	if (hdn_test_insn (opcd, mdrm, &sinstr_table[insn_set].table[ret], 
			   insn, host))
	    goto out;
    }

    /* insn not found */
    bits = 0;
    ret  = 0;

  out:
    if (numbits)
	(*numbits) = bits;
    return ret;
}

/*
 * flips the instruction to the appropriate one, returns the number of
 * bits encoded 
 */
int hdn_flip_insn (struct instr *insn, unsigned char *host,
                   unsigned char *source, int bit_offset)
{
    int insn_set, bits;
    int val = -1, bitmask;
    int num_insns;

#ifdef VARBITS
    do {
#endif
	/*
	 * Get the number of bits we can encode, and using which set of
	 * instructions
	 */
	if (!(bits = _hdn_is_valid_insn (insn, host, val, &insn_set))) return 0;
	num_insns = hdn_num_insns (sinstr_table[insn_set].table);
	
	/* 
	 * create a bitmask for the first n bits 
	 */
	bitmask = ((1 << bits) - 1) << (8 - bits);
	
	/*
	 * Get the next n bits from the source, even if they cross
	 * byte boundaries.  Note: max number of bits we can encode is 8!
	 */
	val = (((source[0] << bit_offset) | (source[1] >> (8 - bit_offset)))
	       & bitmask) >> (8 - bits);
	
	/*
	 * Val now contains the next bits to be encoded.  As we have
	 * limited the range of val, we can use it to index which
	 * instruction to flip to.
	 */

#ifdef VARBITS
	/*
	 * make sure that we really can encode val with the given
	 * number of bits.
	 */
    } while (bits != _hdn_is_valid_insn (insn, host, val, &insn_set));
#endif

    /*
     * do nothing if we already have what we need
     */ 
    if (hdn_insn_val (insn, host, &bits) == val) return bits;

#ifdef VARBITS
    /*
     * this value cannot be expressed, so replaced it with the invalid
     * marker.
     */
    if (((num_insns - pow(2, hdn_log2 (num_insns))) > 2) &&
	val >= (num_insns - 1))
    {
	val = num_insns - 1;
	bits = 0;
    }    
#endif

    /* change the opcode to the appropriate one */
    host[0] = sinstr_table[insn_set].table[val].opcd;

    /* reg bits */
    if (sinstr_table[insn_set].table[val].regval_op == EQ)
    {
	host[1] &= 0xC7;  //clear out the bits
	host[1] |= (sinstr_table[insn_set].table[val].regval << 3) & 0x38;
    }
		
    /* rm bits */
    if (sinstr_table[insn_set].table[val].rmval_op == EQ)
    {
	host[1] &= 0xF8;  //clear out the bits
	host[1] |= (sinstr_table[insn_set].table[val].rmval) & 0x07;
    }
	    
    /* imm val */
    if (sinstr_table[insn_set].table[val].immval_op == EQ)
    {
	hdn_set_imm_val (insn, host, sinstr_table[insn_set].table[val].immval);
    }
    if (sinstr_table[insn_set].table[val].immval_op == NEG)
    {
	hdn_set_imm_val (insn, host, -hdn_get_imm_val (insn, host));
    }
    
    //done
    return bits;
}
